# -*- coding: utf-8 -*-

"""
fairsearchcore.mtable_generator
~~~~~~~~~~~~~~~
Contains the mechanics for creating an mtable
"""

import pandas as pd
import scipy.stats as stats

from fairsearchcore import fail_prob


class MTableGenerator:

    def __init__(self, k, p, alpha, adjust_alpha):
        # assign parameters
        self.k = k
        self.p = p
        self.alpha = alpha
        self.adjust_alpha = adjust_alpha

        if self.adjust_alpha:
            fail_prob_pair = fail_prob.RecursiveNumericFailProbabilityCalculator(k, p, alpha).adjust_alpha()
            self.adjusted_alpha = fail_prob_pair.alpha
            self._mtable = fail_prob_pair.mtable
        else:
            self.adjusted_alpha = alpha
            self._mtable = self._compute_mtable()

    def mtable_as_list(self):
        return [int(i) for i in self._mtable.m.tolist()]

    def mtable_as_dataframe(self):
        return self._mtable

    def m(self, k):
        if k < 1:
            raise ValueError("Parameter k must be at least 1")
        elif k > self.k:
            raise ValueError("Parameter k must be at most {0}".format(self.k))

        result = stats.binom.ppf(self.adjusted_alpha if self.adjust_alpha else self.alpha, k, self.p)
        return 0 if result < 0 else result

    def _compute_mtable(self):
        """ Computes a table containing the minimum number of protected elements
            required at each position
        """
        mtable = pd.DataFrame(columns=["m"])
        for i in range(1, self.k + 1):
            if i % 2000 == 0:
                print("Computing m: {:.0f} of {:.0f}".format(i, self.k))
            mtable.loc[i] = [self.m(i)]
        return mtable


def compute_aux_mtable(mtable):
    """
    Stores the inverse of an mTable entry and the size of the block with respect to the inverse
    """
    if not (isinstance(mtable, pd.DataFrame)):
        raise TypeError("Internal mtable must be a DataFrame")

    aux_mtable = pd.DataFrame(columns=["inv", "block"])
    last_m_seen = 0
    last_position = 0
    for position in range(1, len(mtable)):
        if position % 2000 == 0:
            print("Computing m inverse: {:.0f} of {:.0f}".format(position, len(mtable)))
        if mtable.at[position, "m"] == last_m_seen + 1:
            last_m_seen += 1
            aux_mtable.loc[position] = [position, position - last_position]
            last_position = position
        elif mtable.at[position, "m"] != last_m_seen:
            raise RuntimeError("Inconsistent mtable")

    return aux_mtable



# -*- coding: utf-8 -*-

"""
fairsearchcore.simulator
~~~~~~~~~~~~~~~
This module serves as a wrapper for simulator functionalities
"""

import random
from fairsearchcore import fair
from fairsearchcore import models


def generate_rankings(M, k:int, p):
    """
    Generates M rankings of n elements using Yang-Stoyanovich process
    :param M:           how many rankings to generate
    :param k:           how many elements should each ranking have
    :param p:           what is the probability that a candidate is protected
    :return:            the generated rankings (list of lists of FairScoreDoc))
    """
    rankings = []
    for m in range(M):
        rankings.append(_generate_ranking(k, p))

    return rankings


def compute_fail_probability(rankings, mtable):
    """
    This computes experimentally how many of the M rankings fail to satisfy the mtable
    :param rankings:    rankings that are checked (list of lists of FairScoreDoc)
    :param mtable:      an mtable to check against (list of int)
    :return:            the ratio of failed rankings
    """
    return len(list(filter(lambda x: not fair.check_ranking(x, mtable), rankings))) * 1.0 / len(rankings)


def _generate_ranking(k, p):
    """
    Create a ranking of 'k' positions in which at each position the
        probability that the candidate is protected is 'p'.
    """
    ranking = []
    for i in range(k):
        is_protected = (random.random() <= p)
        ranking.append(models.FairScoreDoc(k-i, k-i, is_protected))
    return ranking



# -*- coding: utf-8 -*-

"""
fairsearchcore.fair
~~~~~~~~~~~~~~~
This module serves as a wrapper around the utilities we have created for FA*IR ranking
"""

import pandas as pd
import warnings

from fairsearchcore import mtable_generator
from fairsearchcore import fail_prob
from fairsearchcore import re_ranker


class Fair:

    k = 10  # the total number of elements
    p = 0.2  # the proportion of protected candidates in the top-k ranking
    alpha = 0.1  # the significance level

    _cache = {} # stores generated mtables in memory

    def __init__(self, k: int, p: float, alpha: float):
        # check the parameters first
        _validate_basic_parameters(k, p, alpha)

        # assign the parameters
        self.k = k
        self.p = p
        self.alpha = alpha

    def create_unadjusted_mtable(self):
        """
        Creates an mtable using alpha unadjusted
        :return:
        """
        return self._create_mtable(self.alpha, False)

    def create_adjusted_mtable(self):
        """
        Creates an mtable using alpha adjusted
        :return:
        """
        return self._create_mtable(self.alpha, True)

    def _create_mtable(self, alpha, adjust_alpha):
        """
        Creates an mtable by using the passed alpha value
        :param alpha:           The significance level
        :param adjust_alpha:    Boolean indicating whether the alpha be adjusted or not
        :return:
        """

        if not (self.k, self.p, self.alpha, adjust_alpha) in self._cache:
            # check if passed alpha is ok
            _validate_alpha(alpha)

            # create the mtable
            fc = mtable_generator.MTableGenerator(self.k, self.p, alpha, adjust_alpha)

            # store as list
            self._cache[(self.k, self.p, self.alpha, adjust_alpha)] = fc.mtable_as_list()

        # return from cache
        return self._cache[(self.k, self.p, self.alpha, adjust_alpha)]

    def adjust_alpha(self) :
        """
        Computes the alpha adjusted for the given set of parameters
        :return:
        """
        rnfpc = fail_prob.RecursiveNumericFailProbabilityCalculator(self.k, self.p, self.alpha)
        fpp = rnfpc.adjust_alpha()
        return fpp.alpha

    def compute_fail_probability(self, mtable):
        """
        Computes analytically the probability that a ranking created with the simulator will fail to pass the mtable
        :return:
        """
        if len(mtable) != self.k:
            raise ValueError("Number of elements k and mtable length must be equal!")

        rnfpc = fail_prob.RecursiveNumericFailProbabilityCalculator(self.k, self.p, self.alpha)

        mtable_df = pd.DataFrame(columns=["m"])

        # transform the list into an pd.DataFrame
        for i in range(1, len(mtable) + 1):
            mtable_df.loc[i] = [mtable[i-1]]

        return rnfpc.calculate_fail_probability(mtable_df)

    def is_fair(self, ranking):
        """
        Checks if the ranking is fair for the given parameters
        :param ranking:     The ranking to be checked (list of FairScoreDoc)
        :return:
        """
        return check_ranking(ranking, self.create_adjusted_mtable())

    def re_rank(self, ranking):
        """
        Applies FA*IR re-ranking to the input ranking with an adjusted mtable
        :param ranking:     The ranking to be re-ranked (list of FairScoreDoc)
        :return:
        """
        return self._re_rank(ranking, True)

    def _re_rank_unadjusted(self, ranking):
        """
        Applies FA*IR re-ranking to the input ranking with an unadjusted mtable
        :param ranking:     The ranking to be re-ranked (list of FairScoreDoc)
        :return:
        """
        return self._re_rank(ranking, False)

    def _re_rank(self, ranking, adjust):
        """
        Applies FA*IR re-ranking to the input ranking and boolean whether to use an adjusted mtable
        :param ranking:     The ranking to be re-ranked (list of FairScoreDoc)
        :return:
        """
        protected = []
        non_protected = []
        for item in ranking:
            if item.is_protected:
                protected.append(item)
            else:
                non_protected.append(item)

        return re_ranker.fair_top_k(self.k, protected, non_protected,
                                    self.create_adjusted_mtable() if adjust else self.create_unadjusted_mtable())


def check_ranking(ranking, mtable):
    """
    Checks if the ranking is fair in respect to the mtable
    :param ranking:     The ranking to be checked (list of FairScoreDoc)
    :param mtable:      The mtable against to check (list of int)
    :return:            Returns whether the rankings satisfies the mtable
    """
    count_protected = 0

    # if the mtable has a different number elements than there are in the top docs return false
    if len(ranking) != len(mtable):
        raise ValueError("Number of documents in ranking and mtable length must be equal!")

    # check number of protected element at each rank
    for i, element in enumerate(ranking):
        count_protected += 1 if element.is_protected else 0
        if count_protected < mtable[i]:
            return False
    return True


def _validate_basic_parameters(k, p, alpha):
    """
    Validates if k, p and alpha are in the required ranges
    :param k:           Total number of elements (above or equal to 10)
    :param p:           The proportion of protected candidates in the top-k ranking (between 0.02 and 0.98)
    :param alpha:       The significance level (between 0.01 and 0.15)
    """
    if k < 10 or k > 400:
        if k < 2:
            raise ValueError("Total number of elements `k` should be between 10 and 400")
        else:
            warnings.warn("Library has not been tested with values outside this range")

    if p < 0.02 or p > 0.98:
        if p < 0 or p > 1:
            raise ValueError("The proportion of protected candidates `p` in the top-k ranking should be between "
                             "0.02 and 0.98")
        else:
            warnings.warn("Library has not been tested with values outside this range")

    _validate_alpha(alpha)


def _validate_alpha(alpha):
    if alpha < 0.01 or alpha > 0.15:
        if alpha < 0.001 or alpha > 0.5:
            raise ValueError("The significance level `alpha` must be between 0.01 and 0.15")
        else:
            warnings.warn("Library has not been tested with values outside this range")




# -*- coding: utf-8 -*-

"""
---------------------------------------
Fair search core for Python
---------------------------------------

This is the Python library with the core algorithms used to do FA*IR ranking.


For details on support methods - see `fairsearchcore.fair` and `fairsearchcore.simulator`. Full documentation
is at https://github.com/fair-search/fairsearch-core/tree/master/python.

:copyright: (c) 2019 by Ivan Kitanovski
:license: Apache 2.0, see LICENSE for more details.
"""
from .fair import check_ranking, Fair
from .simulator import compute_fail_probability, generate_rankings

# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())


# -*- coding: utf-8 -*-

"""
fairsearchcore.models
~~~~~~~~~~~~~~~
This module contains the primary objects that power fairsearchore.
"""


class FairScoreDoc(object):
    """The :class:`FairScoreDoc` object, which is a representation of the items in the rankings.
    Contains a `id`, `score` and `is_protected` attribute
    """

    def __init__(self, id, score, is_protected):
        self.id = id
        self.score = score
        self.is_protected = is_protected

    def __repr__(self):
        return "<FairScoreDoc [%s]>" % ("Protected" if self.is_protected else "Nonprotected")



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='fairsearchcore',
    version='1.0.3',
    description='A Python library with the core algorithms used to do fair search. ',
    long_description=long_description,
    license='Apache 2.0',
    packages=['fairsearchcore'],
    author='Ivan Kitanovski',
    author_email='ivan.kitanovski@gmail.com',
    url='https://github.com/fair-search/fairsearchcore-python',
    keywords=['search','fairness', 'fa*ir', 'ranking', 'reranking'],
    python_requires=">=3.0",
    install_requires=[
        'pandas>=0.23',
        'scipy>=1.1.0',
    ],
    tests_require=[
        'pytest>=2.8.0'
    ],
    setup_requires=['pytest-runner'],
    test_suite="tests",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy'
      ]
)



# -*- coding: utf-8 -*-

"""
fairsearchcore.fail_prob
~~~~~~~~~~~~~~~
Contains the classes to analytically compute fail probability
"""

import abc
from scipy.stats import binom
from fairsearchcore import mtable_generator

EPS = 0.0000000000000001

class FailProbabilityCalculator(abc.ABC):
    """
    Base class for the fail probability calculation
    """

    def __init__(self, k, p, alpha):
        self. k = k
        self.p = p
        self.alpha = alpha

        self.pmf_cache = {}

    @abc.abstractmethod
    def calculate_fail_probability(self, mtable):
        raise NotImplementedError("This is an abstract method. Implement it!")

    def get_from_pmf_cache(self, trials, successes):
        key = (trials, successes)
        if not key in self.pmf_cache:
            # TODO: Check the documentation if this is fine
            self.pmf_cache[key] = binom.pmf(k=successes, n=trials, p=self.p)
        return self.pmf_cache[key]


class RecursiveNumericFailProbabilityCalculator(FailProbabilityCalculator):
    """
    Recursive calculation of fail probability
    """
    def __init__(self, k, p, alpha):
        super().__init__(k, p, alpha)

        self.legal_assignment_cache = {}

    def adjust_alpha(self):
        a_min = 0
        a_max = self.alpha
        a_mid = (a_min + a_max) / 2

        minb = self._compute_boundary(a_min)
        maxb = self._compute_boundary(a_max)
        midb = self._compute_boundary(a_mid)

        while minb.mass_of_mtable() < maxb.mass_of_mtable() and midb.fail_prob != self.alpha:
            if midb.fail_prob < self.alpha:
                a_min = a_mid
                minb = self._compute_boundary(a_min)
            elif midb.fail_prob > self.alpha:
                a_max = a_mid
                maxb = self._compute_boundary(a_max)

            a_mid = (a_min + a_max) / 2
            midb = self._compute_boundary(a_mid)

            max_mass = maxb.mass_of_mtable()
            min_mass = minb.mass_of_mtable()
            mid_mass = midb.mass_of_mtable()

            if max_mass - min_mass == 1 or maxb.alpha - minb.alpha <= EPS:
                min_diff = abs(minb.fail_prob - self.alpha)
                max_diff = abs(maxb.fail_prob - self.alpha)

                if min_diff <= max_diff:
                    return minb
                else:
                    return maxb

            if max_mass - mid_mass == 1 and mid_mass - min_mass == 1:
                min_diff = abs(minb.fail_prob - self.alpha)
                max_diff = abs(maxb.fail_prob - self.alpha)
                mid_diff = abs(midb.fail_prob - self.alpha)

                if mid_diff <= max_diff and mid_diff <= min_diff:
                    return midb
                if min_diff <= mid_diff and min_diff <= max_diff:
                    return minb
                else:
                    return maxb

        return midb

    def calculate_fail_probability(self, mtable):
        """
        Analytically calculates the fail probability of the mtable
        """
        aux_mtable = mtable_generator.compute_aux_mtable(mtable)
        max_protected = aux_mtable['block'].sum()
        block_sizes = aux_mtable['block'].tolist()#[1:]
        success_prob = self._find_legal_assignments(max_protected, block_sizes)
        return 0 if success_prob == 0 else (1 - success_prob)

    def _compute_boundary(self, alpha):
        """
        Returns a tuple of (k, p, alpha, fail_prob, mtable)
        """
        mtable = mtable_generator.MTableGenerator(self.k, self.p, alpha, False).mtable_as_dataframe()
        fail_prob = self.calculate_fail_probability(mtable)
        return MTableFailProbPair(self.k, self.p, alpha, fail_prob, mtable)

    def _find_legal_assignments(self, number_of_candidates, block_sizes):
        return self._find_legal_assignments_aux(number_of_candidates, block_sizes, 1, 0)

    def _find_legal_assignments_aux(self, number_of_candidates, block_sizes,
                                    current_block_number, candidates_assigned_so_far):
        if len(block_sizes) == 0:
            return 1

        min_needed_this_block = current_block_number - candidates_assigned_so_far
        if min_needed_this_block < 0:
            min_needed_this_block = 0

        max_possible_this_block = min(block_sizes[0], number_of_candidates)

        assignments = 0
        new_remaining_block_sizes = block_sizes[1:]
        for items_this_block in range(min_needed_this_block, max_possible_this_block + 1):
            new_remaining_candidates = number_of_candidates - items_this_block

            suffixes = self._calculate_legal_assignments_aux(new_remaining_candidates, new_remaining_block_sizes,
                                                             current_block_number + 1,
                                                             candidates_assigned_so_far + items_this_block)

            assignments += self.get_from_pmf_cache(max_possible_this_block, items_this_block) * suffixes

        return assignments

    def _calculate_legal_assignments_aux(self, remaining_candidates, remaining_block_sizes,
                                         current_block_number, candidates_assigned_so_far):
        key = LegalAssignmentKey(remaining_candidates, remaining_block_sizes,
                                 current_block_number, candidates_assigned_so_far)

        if not key in self.legal_assignment_cache:
            self.legal_assignment_cache[key] = self._find_legal_assignments_aux(remaining_candidates,
                                                                                remaining_block_sizes,
                                                                                current_block_number,
                                                                                candidates_assigned_so_far)

        return self.legal_assignment_cache[key]


class LegalAssignmentKey:
    """
    Utility class for the recursive fail prob
    """

    def __init__(self, remaining_candidates, remaining_block_sizes, current_block_number, candidates_assigned_so_far):
        self.remaining_candidates = remaining_candidates
        self.remaining_block_sizes = remaining_block_sizes
        self.current_block_number = current_block_number
        self.candidates_assigned_so_far = candidates_assigned_so_far

    def __eq__(self, other):
        if self.remaining_candidates != other.remaining_candidates:
            return False
        if self.current_block_number != other.current_block_number:
            return False
        if self.candidates_assigned_so_far != other.candidates_assigned_so_far:
            return False
        if self.remaining_block_sizes != other.remaining_block_sizes:
            return False
        return True

    def __hash__(self):
        return (self.remaining_candidates + len(self.remaining_block_sizes) << 16) \
               + self.current_block_number + self.candidates_assigned_so_far


class MTableFailProbPair:
    """
    Encapsulation of all parameters for the interim mtables
    """
    def __init__(self, k, p, alpha, fail_prob, mtable):
        self.k = k
        self.p = p
        self.alpha = alpha
        self.fail_prob = fail_prob
        self.mtable = mtable

    def mass_of_mtable(self):
        return self.mtable['m'].sum()



"""
fairsearchcore.re_ranker
~~~~~~~~~~~~~~~
This module contains the FA*IR re-ranking algorithm
"""


def fair_top_k(k, protected_candidates, non_protected_candidates, mtable):
    """    
    Parameters:
    ----------
    k : int
        the expected length of the ranking
    protected_candidates : [FairScoreDoc]
        array of protected class:`candidates <fairsearhcore.models.FairScoreDoc>`, assumed to be
        sorted by item score in descending order
    non_protected_candidates : [FairScoreDoc]
        array of non-protected class:`candidates <fairsearhcore.models.FairScoreDoc>`, assumed to be
        sorted by item score in descending order
        significance level for the binomial cumulative distribution function -> minimum probability at
        which a fair ranking contains the minProp amount of protected candidates
    Return:
    ------
    an array of elements in the form of `dict` that maximizes ordering and
    selection fairness
    the left-over candidates that were not selected into the ranking, sorted color-blindly
    """

    result = []
    countProtected = 0

    idxProtected = 0
    idxNonProtected = 0

    for i in range(k):
        if idxProtected >= len(protected_candidates) and idxNonProtected >= len(non_protected_candidates):
            # no more candidates available, return list shorter than k
            return result, []
        if idxProtected >= len(protected_candidates):
            # no more protected candidates available, take non-protected instead
            result.append(non_protected_candidates[idxNonProtected])
            idxNonProtected += 1

        elif idxNonProtected >= len(non_protected_candidates):
            # no more non-protected candidates available, take protected instead
            result.append(protected_candidates[idxProtected])
            idxProtected += 1
            countProtected += 1
        elif countProtected < mtable[i]:
            # add a protected candidate
            result.append(protected_candidates[idxProtected])
            idxProtected += 1
            countProtected += 1
        else:
            # find the best candidate available
            if protected_candidates[idxProtected].score >= non_protected_candidates[idxNonProtected].score:
                # the best is a protected one
                result.append(protected_candidates[idxProtected])
                idxProtected += 1
                countProtected += 1
            else:
                # the best is a non-protected one
                result.append(non_protected_candidates[idxNonProtected])
                idxNonProtected += 1

    return result # , __mergeTwoRankings(protected_candidates[idxProtected:], non_protected_candidates[idxNonProtected:])


def __mergeTwoRankings(ranking1, ranking2):
    result = ranking1 + ranking2
    result.sort(key=lambda candidate: candidate.score, reverse=True)
    return result